@@ -19,6 +19,8 @@
 
 #include <sys/types.h>
 #include <sys/uio.h>
+#include <sys/socket.h>
+#include <netinet/in.h>
 #include <arpa/nameser.h>
 #ifdef YP
 #include <rpc/rpc.h>
@@ -29,6 +31,7 @@
 
 #include <err.h>
 #include <errno.h>
+#include <resolv.h> /* for res_hnok */
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
@@ -45,6 +48,7 @@ struct match {
 static int getaddrinfo_async_run(struct async *, struct async_res *);
 static int get_port(const char *, const char *, int);
 static int iter_family(struct async *, int);
+static int iter_domain(struct async *, const char *, char *, size_t);
 static int addrinfo_add(struct async *, const struct sockaddr *, const char *);
 static int addrinfo_from_file(struct async *, int,  FILE *);
 static int addrinfo_from_pkt(struct async *, char *, size_t);
@@ -68,20 +72,31 @@ static const struct match matches[] = {
 #define MATCH_SOCKTYPE(a, b) ((a) == matches[(b)].socktype || ((a) == 0 && \
 				matches[(b)].socktype != SOCK_RAW))
 
+enum {
+	DOM_INIT,
+	DOM_DOMAIN,
+	DOM_DONE
+};
+
 struct async *
 getaddrinfo_async(const char *hostname, const char *servname,
 	const struct addrinfo *hints, struct asr *asr)
 {
 	struct asr_ctx	*ac;
 	struct async	*as;
+	char		 alias[MAXDNAME];
 
 	ac = asr_use_resolver(asr);
 	if ((as = asr_async_new(ac, ASR_GETADDRINFO)) == NULL)
 		goto abort; /* errno set */
 	as->as_run = getaddrinfo_async_run;
 
-	if (hostname && (as->as.ai.hostname = strdup(hostname)) == NULL)
-		goto abort; /* errno set */
+	if (hostname) {
+		if (asr_hostalias(ac, hostname, alias, sizeof(alias)))
+			hostname = alias;
+		if ((as->as.ai.hostname = strdup(hostname)) == NULL)
+			goto abort; /* errno set */
+	}
 	if (servname && (as->as.ai.servname = strdup(servname)) == NULL)
 		goto abort; /* errno set */
 	if (hints)
@@ -107,8 +122,9 @@ getaddrinfo_async_run(struct async *as, struct async_res *ar)
 	static char	*domain = NULL;
 	char		*res;
 	int		 len;
-	char		 alias[MAXDNAME], *name;
+	char		 *name;
 #endif
+	char		 fqdn[MAXDNAME];
 	const char	*str;
 	struct addrinfo	*ai;
 	int		 i, family, r;
@@ -290,30 +306,60 @@ getaddrinfo_async_run(struct async *as, struct async_res *ar)
 				ar->ar_gai_errno = 0;
 				async_set_state(as, ASR_STATE_HALT);
 			} else
-				async_set_state(as, ASR_STATE_NEXT_DB);
+				async_set_state(as, ASR_STATE_NEXT_DOMAIN);
 			break;
 		}
 		async_set_state(as, ASR_STATE_SAME_DB);
 		break;
 
+	case ASR_STATE_NEXT_DOMAIN:
+		/* domain search is only for dns */
+		if (AS_DB(as) != ASR_DB_DNS) {
+			async_set_state(as, ASR_STATE_NEXT_DB);
+			break;
+		}
+		as->as_family_idx = 0;
+
+		free(as->as.ai.fqdn);
+		as->as.ai.fqdn = NULL;
+		r = iter_domain(as, as->as.ai.hostname, fqdn, sizeof(fqdn));
+		if (r == -1) {
+			async_set_state(as, ASR_STATE_NEXT_DB);
+			break;
+		}
+		if (r == 0) {
+			ar->ar_gai_errno = EAI_FAIL;
+			async_set_state(as, ASR_STATE_HALT);
+			break;
+		}
+		as->as.ai.fqdn = strdup(fqdn);
+		if (as->as.ai.fqdn == NULL) {
+			ar->ar_gai_errno = EAI_MEMORY;
+			async_set_state(as, ASR_STATE_HALT);
+		}
+
+		async_set_state(as, ASR_STATE_SAME_DB);
+		break;
+
 	case ASR_STATE_SAME_DB:
 		/* query the current DB again */
 		switch (AS_DB(as)) {
 		case ASR_DB_DNS:
+			if (as->as.ai.fqdn == NULL) {
+				/* First try, initialize domain iteration */
+				as->as_dom_flags = 0;
+				as->as_dom_step = DOM_INIT;
+				async_set_state(as, ASR_STATE_NEXT_DOMAIN);
+				break;
+			}
+
 			family = (as->as.ai.hints.ai_family == AF_UNSPEC) ?
 			    AS_FAMILY(as) : as->as.ai.hints.ai_family;
-			if (as->as.ai.fqdn) {
-				as->as.ai.subq = res_query_async_ctx(
-				    as->as.ai.fqdn, C_IN,
-				    (family == AF_INET6) ? T_AAAA : T_A,
-				    as->as_ctx);
-			}
-			else {
-				as->as.ai.subq = res_search_async_ctx(
-				    as->as.ai.hostname, C_IN,
-				    (family == AF_INET6) ? T_AAAA : T_A,
-				    as->as_ctx);
-			}
+
+			as->as.ai.subq = res_query_async_ctx(as->as.ai.fqdn,
+			    C_IN, (family == AF_INET6) ? T_AAAA : T_A,
+			    as->as_ctx);
+
 			if (as->as.ai.subq == NULL) {
 				if (errno == ENOMEM)
 					ar->ar_gai_errno = EAI_MEMORY;
@@ -355,10 +401,7 @@ getaddrinfo_async_run(struct async *as, struct async_res *ar)
 			family = (as->as.ai.hints.ai_family == AF_UNSPEC) ?
 			    AS_FAMILY(as) : as->as.ai.hints.ai_family;
 
-			name = asr_hostalias(as->as_ctx, as->as.ai.hostname,
-			    alias, sizeof(alias));
-			if (name == NULL)
-				name = as->as.ai.hostname;
+			name = as->as.ai.hostname;
 
 			/* XXX
 			 * ipnodes.byname could also contain IPv4 address
@@ -504,6 +547,115 @@ iter_family(struct async *as, int first)
 	return AS_FAMILY(as);
 }
 
+/*
+ * Concatenate a name and a domain name. The result has no trailing dot.
+ * Return the resulting string length, or 0 in case of error.
+ */
+static size_t
+domcat(const char *name, const char *domain, char *buf, size_t buflen)
+{
+	size_t	r;
+
+	r = asr_make_fqdn(name, domain, buf, buflen);
+	if (r == 0)
+		return (0);
+	buf[r - 1] = '\0';
+
+	return (r - 1);
+}
+
+/*
+ * Implement the search domain strategy.
+ *
+ * XXX duplicate from res_search_async
+ *
+ * This function works as a generator that constructs complete domains in
+ * buffer "buf" of size "len" for the given host name "name", according to the
+ * search rules defined by the resolving context.  It is supposed to be called
+ * multiple times (with the same name) to generate the next possible domain
+ * name, if any.
+ *
+ * It returns -1 if all possibilities have been exhausted, 0 if there was an
+ * error generating the next name, or the resulting name length.
+ */
+static int
+iter_domain(struct async *as, const char *name, char * buf, size_t len)
+{
+	const char	*c;
+	int		 dots;
+
+	switch (as->as_dom_step) {
+
+	case DOM_INIT:
+		/* First call */
+
+		/*
+		 * If "name" is an FQDN, that's the only result and we
+		 * don't try anything else.
+		 */
+		if (strlen(name) && name[strlen(name) - 1] ==  '.') {
+			DPRINT("asr: iter_domain(\"%s\") fqdn\n", name);
+			as->as_dom_flags |= ASYNC_DOM_FQDN;
+			as->as_dom_step = DOM_DONE;
+			return (domcat(name, NULL, buf, len));
+		}
+
+		/*
+		 * Otherwise, we iterate through the specified search domains.
+		 */
+		as->as_dom_step = DOM_DOMAIN;
+		as->as_dom_idx = 0;
+
+		/*
+		 * If "name" as enough dots, use it as-is first, as indicated
+		 * in resolv.conf(5).
+		 */
+		dots = 0;
+		for (c = name; *c; c++)
+			dots += (*c == '.');
+		if (dots >= as->as_ctx->ac_ndots) {
+			DPRINT("asr: iter_domain(\"%s\") ndots\n", name);
+			as->as_dom_flags |= ASYNC_DOM_NDOTS;
+			if (strlcpy(buf, name, len) >= len)
+				return (0);
+			return (strlen(buf));
+		}
+		/* Otherwise, starts using the search domains */
+		/* FALLTHROUGH */
+
+	case DOM_DOMAIN:
+		if (as->as_dom_idx < as->as_ctx->ac_domcount) {
+			DPRINT("asr: iter_domain(\"%s\") domain \"%s\"\n",
+			    name, as->as_ctx->ac_dom[as->as_dom_idx]);
+			as->as_dom_flags |= ASYNC_DOM_DOMAIN;
+			return (domcat(name,
+			    as->as_ctx->ac_dom[as->as_dom_idx++], buf, len));
+		}
+
+		/* No more domain to try. */
+
+		as->as_dom_step = DOM_DONE;
+
+		/*
+		 * If the name was not tried as an absolute name before,
+		 * do it now.
+		 */
+		if (!(as->as_dom_flags & ASYNC_DOM_NDOTS)) {
+			DPRINT("asr: iter_domain(\"%s\") as is\n", name);
+			as->as_dom_flags |= ASYNC_DOM_ASIS;
+			if (strlcpy(buf, name, len) >= len)
+				return (0);
+			return (strlen(buf));
+		}
+		/* Otherwise, we are done. */
+
+	case DOM_DONE:
+	default:
+		DPRINT("asr: iter_domain(\"%s\") done\n", name);
+		return (-1);
+	}
+}
+
 /*
  * Use the sockaddr at "sa" to extend the result list on the "as" context,
  * with the specified canonical name "cname". This function adds one
@@ -588,7 +740,7 @@ asr_freeaddrinfo(struct addrinfo *ai)
 static int
 addrinfo_from_file(struct async *as, int family, FILE *f)
 {
-	char		*tokens[MAXTOKEN], buf[MAXDNAME], *name, *c;
+	char		*tokens[MAXTOKEN], *c;
 	int		 n, i;
 	union {
 		struct sockaddr		sa;
@@ -596,17 +748,13 @@ addrinfo_from_file(struct async *as, int family, FILE *f)
 		struct sockaddr_in6	sain6;
 	} u;
 
-	name = asr_hostalias(as->as_ctx, as->as.ai.hostname, buf, sizeof(buf));
-	if (name == NULL)
-		name = as->as.ai.hostname;
-
 	for (;;) {
 		n = asr_parse_namedb_line(f, tokens, MAXTOKEN);
 		if (n == -1)
 			break; /* ignore errors reading the file */
 
 		for (i = 1; i < n; i++) {
-			if (strcasecmp(name, tokens[i]))
+			if (strcasecmp(as->as.ai.hostname, tokens[i]))
 				continue;
 			if (asr_sockaddr_from_str(&u.sa, family, tokens[0]) == -1)
 				continue;
@@ -652,14 +800,6 @@ addrinfo_from_pkt(struct async *as, char *pkt, size_t pktlen)
 		    rr.rr_class != q.q_class)
 			continue;
 
-		if (as->as.ai.fqdn == NULL) {
-			asr_strdname(q.q_dname, buf, sizeof buf);
-			buf[strlen(buf) - 1] = '\0';
-			as->as.ai.fqdn = strdup(buf);
-			if (as->as.ai.fqdn == NULL)
-				return (-1); /* errno set */
-		}
-
 		memset(&u, 0, sizeof u);
 		if (rr.rr_type == T_A) {
 #ifdef HAVE_STRUCT_SOCKADDR_IN_SIN_LEN
@@ -681,7 +821,7 @@ addrinfo_from_pkt(struct async *as, char *pkt, size_t pktlen)
 		if (as->as.ai.hints.ai_flags & AI_CANONNAME) {
 			asr_strdname(rr.rr_dname, buf, sizeof buf);
 			buf[strlen(buf) - 1] = '\0';
-			c = buf;
+			c = res_hnok(buf) ? buf : NULL;
 		} else if (as->as.ai.hints.ai_flags & AI_FQDN)
 			c = as->as.ai.fqdn;
 		else
