@@ -69,6 +69,18 @@ envelope_set_errormsg(struct envelope *e, char *fmt, ...)
 		strlcpy(e->errorline + (sizeof(e->errorline) - 4), "...", 4);
 }
 
+void
+envelope_set_esc_class(struct envelope *e, enum enhanced_status_class class)
+{
+	e->esc_class = class;
+}
+
+void
+envelope_set_esc_code(struct envelope *e, enum enhanced_status_code code)
+{
+	e->esc_code = code;
+}
+
 static int
 envelope_buffer_to_dict(struct dict *d,  const char *ibuf, size_t buflen)
 {
@@ -76,7 +88,7 @@ envelope_buffer_to_dict(struct dict *d,  const char *ibuf, size_t buflen)
 	size_t		 len;
 	char		*buf, *field, *nextline;
 
-	bzero(lbuf, sizeof lbuf);
+	memset(lbuf, 0, sizeof lbuf);
 	if (strlcpy(lbuf, ibuf, sizeof lbuf) >= sizeof lbuf)
 		goto err;
 	buf = lbuf;
@@ -158,7 +170,7 @@ envelope_load_buffer(struct envelope *ep, const char *ibuf, size_t buflen)
 		goto end;
 	}
 
-	bzero(ep, sizeof *ep);
+	memset(ep, 0, sizeof *ep);
 	ret = envelope_ascii_load(ep, &d);
 	if (ret)
 		ep->version = SMTPD_ENVELOPE_VERSION;
@@ -190,6 +202,12 @@ envelope_dump_buffer(const struct envelope *ep, char *dest, size_t len)
 	envelope_ascii_dump(ep, &dest, &len, "expire");
 	envelope_ascii_dump(ep, &dest, &len, "retry");
 	envelope_ascii_dump(ep, &dest, &len, "flags");
+	envelope_ascii_dump(ep, &dest, &len, "dsn-notify");
+	envelope_ascii_dump(ep, &dest, &len, "dsn-ret");
+	envelope_ascii_dump(ep, &dest, &len, "dsn-envid");
+	envelope_ascii_dump(ep, &dest, &len, "dsn-orcpt");
+	envelope_ascii_dump(ep, &dest, &len, "esc-class");
+	envelope_ascii_dump(ep, &dest, &len, "esc-code");
 
 	switch (ep->type) {
 	case D_MDA:
@@ -222,6 +240,17 @@ envelope_dump_buffer(const struct envelope *ep, char *dest, size_t len)
 	return (dest - p);
 }
 
+static int
+ascii_load_uint8(uint8_t *dest, char *buf)
+{
+	const char *errstr;
+
+	*dest = strtonum(buf, 0, 0xff, &errstr);
+	if (errstr)
+		return 0;
+	return 1;
+}
+
 static int
 ascii_load_uint16(uint16_t *dest, char *buf)
 {
@@ -283,8 +312,8 @@ ascii_load_sockaddr(struct sockaddr_storage *ss, char *buf)
 	struct sockaddr_in6 ssin6;
 	struct sockaddr_in  ssin;
 
-	bzero(&ssin, sizeof ssin);
-	bzero(&ssin6, sizeof ssin6);
+	memset(&ssin, 0, sizeof ssin);
+	memset(&ssin6, 0, sizeof ssin6);
 
 	if (!strcmp("local", buf)) {
 		ss->ss_family = AF_LOCAL;
@@ -388,6 +417,20 @@ ascii_load_bounce_type(enum bounce_type *dest, char *buf)
 		*dest = B_ERROR;
 	else if (strcasecmp(buf, "warn") == 0)
 		*dest = B_WARNING;
+	else if (strcasecmp(buf, "dsn") == 0)
+		*dest = B_DSN;
+	else
+		return 0;
+	return 1;
+}
+
+static int
+ascii_load_dsn_ret(enum dsn_ret *ret, char *buf)
+{
+	if (strcasecmp(buf, "HDRS") == 0)
+		*ret = DSN_RETHDRS;
+	else if (strcasecmp(buf, "FULL") == 0)
+		*ret = DSN_RETFULL;
 	else
 		return 0;
 	return 1;
@@ -464,8 +507,8 @@ ascii_load_field(const char *field, struct envelope *ep, char *buf)
 		    sizeof ep->agent.mta.relay.authtable);
 
 	if (strcasecmp("mta-relay-cert", field) == 0)
-		return ascii_load_string(ep->agent.mta.relay.cert, buf,
-		    sizeof ep->agent.mta.relay.cert);
+		return ascii_load_string(ep->agent.mta.relay.pki_name, buf,
+		    sizeof ep->agent.mta.relay.pki_name);
 
 	if (strcasecmp("mta-relay-flags", field) == 0)
 		return ascii_load_mta_relay_flags(&ep->agent.mta.relay.flags, buf);
@@ -506,6 +549,24 @@ ascii_load_field(const char *field, struct envelope *ep, char *buf)
 	if (strcasecmp("version", field) == 0)
 		return ascii_load_uint32(&ep->version, buf);
 
+	if (strcasecmp("dsn-notify", field) == 0)
+		return ascii_load_uint8(&ep->dsn_notify, buf);
+
+	if (strcasecmp("dsn-orcpt", field) == 0)
+		return ascii_load_mailaddr(&ep->dsn_orcpt, buf);
+
+	if (strcasecmp("dsn-ret", field) == 0)
+		return ascii_load_dsn_ret(&ep->dsn_ret, buf);
+
+	if (strcasecmp("dsn-envid", field) == 0)
+		return ascii_load_string(ep->dsn_envid, buf, sizeof(ep->dsn_envid));
+
+	if (strcasecmp("esc-class", field) == 0)
+		return ascii_load_uint8(&ep->esc_class, buf);
+
+	if (strcasecmp("esc-code", field) == 0)
+		return ascii_load_uint8(&ep->esc_code, buf);
+
 	return (0);
 }
 
@@ -529,6 +590,12 @@ err:
 }
 
 
+static int
+ascii_dump_uint8(uint8_t src, char *dest, size_t len)
+{
+	return bsnprintf(dest, len, "%d", src);
+}
+
 static int
 ascii_dump_uint16(uint16_t src, char *dest, size_t len)
 {
@@ -673,12 +740,30 @@ ascii_dump_bounce_type(enum bounce_type type, char *dest, size_t len)
 	case B_WARNING:
 		p = "warn";
 		break;
+	case B_DSN:
+		p = "dsn";
+		break;
 	default:
 		return 0;
 	}
 	return bsnprintf(dest, len, "%s", p);
 }
 
+
+static int
+ascii_dump_dsn_ret(enum dsn_ret flag, char *dest, size_t len)
+{
+        size_t cpylen = 0;
+
+        dest[0] = '\0';
+        if (flag == DSN_RETFULL)
+                cpylen = strlcat(dest, "FULL", len);
+        else if (flag == DSN_RETHDRS)
+                cpylen = strlcat(dest, "HDRS", len);
+
+        return cpylen < len ? 1 : 0;
+}
+
 static int
 ascii_dump_field(const char *field, const struct envelope *ep,
     char *buf, size_t len)
@@ -748,7 +833,7 @@ ascii_dump_field(const char *field, const struct envelope *ep,
 		    buf, len);
 
 	if (strcasecmp(field, "mta-relay-cert") == 0)
-		return ascii_dump_string(ep->agent.mta.relay.cert,
+		return ascii_dump_string(ep->agent.mta.relay.pki_name,
 		    buf, len);
 
 	if (strcasecmp(field, "mta-relay-flags") == 0)
@@ -791,6 +876,33 @@ ascii_dump_field(const char *field, const struct envelope *ep,
 	if (strcasecmp(field, "version") == 0)
 		return ascii_dump_uint32(SMTPD_ENVELOPE_VERSION, buf, len);
 
+	if (strcasecmp(field, "dsn-notify") == 0)
+		return ascii_dump_uint8(ep->dsn_notify, buf, len);
+
+	if (strcasecmp(field, "dsn-ret") == 0)
+		return ascii_dump_dsn_ret(ep->dsn_ret, buf, len);
+
+	if (strcasecmp(field, "dsn-orcpt") == 0) {
+		if (ep->dsn_orcpt.user[0] && ep->dsn_orcpt.domain[0])
+			return ascii_dump_mailaddr(&ep->dsn_orcpt, buf, len);
+		return 1;
+	}
+
+	if (strcasecmp(field, "dsn-envid") == 0)
+		return ascii_dump_string(ep->dsn_envid, buf, len);
+
+	if (strcasecmp(field, "esc-class") == 0) {
+		if (ep->esc_class)
+			return ascii_dump_uint8(ep->esc_class, buf, len);
+		return 1;
+	}
+
+	if (strcasecmp(field, "esc-code") == 0) {
+		if (ep->esc_class)
+			return ascii_dump_uint8(ep->esc_code, buf, len);
+		return 1;
+	}
+
 	return (0);
 }
 
@@ -803,7 +915,7 @@ envelope_ascii_dump(const struct envelope *ep, char **dest, size_t *len, const c
 	if (*dest == NULL)
 		return;
 
-	bzero(buf, sizeof buf);
+	memset(buf, 0, sizeof buf);
 	if (! ascii_dump_field(field, ep, buf, sizeof buf))
 		goto err;
 	if (buf[0] == '\0')
