@@ -416,7 +416,7 @@ filter_api_init(void)
 	tree_init(&sessions);
 	event_init();
 
-	bzero(&fi, sizeof(fi));
+	memset(&fi, 0, sizeof(fi));
 	fi.p.proc = PROC_MFA;
 	fi.p.name = "filter";
 	fi.p.handler = filter_dispatch;
