@@ -87,7 +87,7 @@ struct modifiers {
 	{ "strip",	mod_strip },
 	{ "raw",	NULL },		/* special case, must stay last */
 };
-static const char	*unsafe = "*?";
+static const char	*unsafe = MAILADDR_ESCAPE;
 
 
 static int		init;
@@ -115,7 +115,7 @@ lka_session(uint64_t id, struct envelope *envelope)
 	lks->envelope = *envelope;
 
 	TAILQ_INIT(&lks->nodes);
-	bzero(&xn, sizeof xn);
+	memset(&xn, 0, sizeof xn);
 	xn.type = EXPAND_ADDRESS;
 	xn.u.mailaddr = lks->envelope.rcpt;
 	lks->expand.rule = NULL;
@@ -346,7 +346,7 @@ lka_expand(struct lka_session *lks, struct rule *rule, struct expandnode *xn)
 			lks->expand.rule = rule;
 			lks->expand.parent = xn;
 			lks->expand.alias = 1;
-			bzero(&node, sizeof node);
+			memset(&node, 0, sizeof node);
 			node.type = EXPAND_USERNAME;
 			mailaddr_to_username(&xn->u.mailaddr, node.u.user,
 				sizeof node.u.user);
@@ -410,7 +410,7 @@ lka_expand(struct lka_session *lks, struct rule *rule, struct expandnode *xn)
 		lks->rule = rule;
 		lks->node = xn;
 
-		bzero(&fwreq, sizeof(fwreq));
+		memset(&fwreq, 0, sizeof(fwreq));
 		fwreq.id = lks->id;
 		(void)strlcpy(fwreq.user, lk.userinfo.username, sizeof(fwreq.user));
 		(void)strlcpy(fwreq.directory, lk.userinfo.directory, sizeof(fwreq.directory));
@@ -551,7 +551,7 @@ lka_submit(struct lka_session *lks, struct rule *rule, struct expandnode *xn)
 			strlcpy(ep->agent.mda.buffer, rule->r_value.buffer,
 			    sizeof ep->agent.mda.buffer);
 
-			bzero(tag, sizeof tag);
+			memset(tag, 0, sizeof tag);
 			if (! mailaddr_tag(&ep->dest, tag, sizeof tag)) {
 				lks->error = LKA_PERMFAIL;
 				free(ep);
@@ -707,7 +707,7 @@ lka_expand_token(char *dest, size_t len, const char *token,
 		} while ((mods = sep) != NULL);
 	}
 	
-	if (! raw)
+	if (! raw && replace)
 		for (i = 0; (size_t)i < strlen(tmp); ++i)
 			if (strchr(unsafe, tmp[i]))
 				tmp[i] = ':';
@@ -768,7 +768,7 @@ lka_expand_format(char *buf, size_t len, const struct envelope *ep,
 	if (len < sizeof tmpbuf)
 		fatalx("lka_expand_format: tmp buffer < rule buffer");
 
-	bzero(tmpbuf, sizeof tmpbuf);
+	memset(tmpbuf, 0, sizeof tmpbuf);
 	pbuf = buf;
 	ptmp = tmpbuf;
 	ret = tmpret = 0;
