@@ -99,7 +99,6 @@ enum mta_state {
 #define MTA_EXT_AUTH_PLAIN     	0x08
 #define MTA_EXT_AUTH_LOGIN     	0x10
 
-
 struct mta_session {
 	uint64_t		 id;
 	struct mta_relay	*relay;
@@ -128,9 +127,7 @@ struct mta_session {
 	struct mta_envelope	*currevp;
 	FILE			*datafp;
 
-#define	MAX_FAILED_ENVELOPES	15
-	struct mta_envelope	*failed[MAX_FAILED_ENVELOPES];
-	int			 failedcount;
+	size_t			 failures;
 };
 
 static void mta_session_init(void);
@@ -151,7 +148,9 @@ static int mta_check_loop(FILE *);
 static void mta_start_tls(struct mta_session *);
 static int mta_verify_certificate(struct mta_session *);
 static struct mta_session *mta_tree_pop(struct tree *, uint64_t);
-static void mta_flush_failedqueue(struct mta_session *);
+static const char * dsn_strret(enum dsn_ret);
+static const char * dsn_strnotify(uint8_t);
+
 void mta_hoststat_update(const char *, const char *);
 void mta_hoststat_reschedule(const char *);
 void mta_hoststat_cache(const char *, uint64_t);
@@ -197,7 +196,7 @@ mta_session(struct mta_relay *relay, struct mta_route *route)
 
 	if (relay->flags & RELAY_SSL && relay->flags & RELAY_AUTH)
 		s->flags |= MTA_USE_AUTH;
-	if (relay->cert)
+	if (relay->pki_name)
 		s->flags |= MTA_USE_CERT;
 	if (relay->flags & RELAY_LMTP)
 		s->flags |= MTA_LMTP;
@@ -320,7 +319,7 @@ mta_session_imsg(struct mproc *p, struct imsg *imsg)
 			return;
 
 		if (resp_ca_cert->status == CA_FAIL) {
-			if (s->relay->cert) {
+			if (s->relay->pki_name) {
 				log_info("smtp-out: Disconnecting session %016"PRIx64
 				    ": CA failure", s->id);
 				mta_free(s);
@@ -335,10 +334,7 @@ mta_session_imsg(struct mproc *p, struct imsg *imsg)
 			}
 		}
 
-		resp_ca_cert = xmemdup(imsg->data, sizeof *resp_ca_cert,
-		    "mta:ca_cert");
-		if (resp_ca_cert == NULL)
-			fatal(NULL);
+		resp_ca_cert = xmemdup(imsg->data, sizeof *resp_ca_cert, "mta:ca_cert");
 		resp_ca_cert->cert = xstrdup((char *)imsg->data +
 		    sizeof *resp_ca_cert, "mta:ca_cert");
 		resp_ca_cert->key = xstrdup((char *)imsg->data +
@@ -350,8 +346,8 @@ mta_session_imsg(struct mproc *p, struct imsg *imsg)
 			fatal("mta: ssl_mta_init");
 		io_start_tls(&s->io, ssl);
 
-		bzero(resp_ca_cert->cert, resp_ca_cert->cert_len);
-		bzero(resp_ca_cert->key, resp_ca_cert->key_len);
+		memset(resp_ca_cert->cert, 0, resp_ca_cert->cert_len);
+		memset(resp_ca_cert->key, 0, resp_ca_cert->key_len);
 		free(resp_ca_cert->cert);
 		free(resp_ca_cert->key);
 		free(resp_ca_cert);
@@ -437,8 +433,6 @@ mta_free(struct mta_session *s)
 		runq_cancel(hangon, NULL, s);
 	}
 
-	mta_flush_failedqueue(s);
-
 	io_clear(&s->io);
 	iobuf_clear(&s->iobuf);
 
@@ -586,6 +580,8 @@ mta_connect(struct mta_session *s)
 static void
 mta_enter_state(struct mta_session *s, int newstate)
 {
+	struct mta_envelope	 *e;
+	size_t			 envid_sz;
 	int			 oldstate;
 	ssize_t			 q;
 	char			 ibuf[SMTPD_MAXLINESIZE];
@@ -678,36 +674,38 @@ mta_enter_state(struct mta_session *s, int newstate)
 		break;
 
 	case MTA_AUTH_LOGIN_USER:
-		bzero(ibuf, sizeof ibuf);
-		if (__b64_pton(s->relay->secret, (unsigned char *)ibuf, sizeof(ibuf)-1) == -1) {
+		memset(ibuf, 0, sizeof ibuf);
+		if (base64_decode(s->relay->secret, (unsigned char *)ibuf,
+				  sizeof(ibuf)-1) == -1) {
 			log_debug("debug: mta: %p: credentials too large on session", s);
 			mta_error(s, "Credentials too large");
 			break;
 		}
 
-		bzero(obuf, sizeof obuf);
-		__b64_ntop((unsigned char *)ibuf + 1, strlen(ibuf + 1), obuf, sizeof obuf);
+		memset(obuf, 0, sizeof obuf);
+		base64_encode((unsigned char *)ibuf + 1, strlen(ibuf + 1), obuf, sizeof obuf);
 		mta_send(s, "%s", obuf);
 
-		bzero(ibuf, sizeof ibuf);
-		bzero(obuf, sizeof obuf);
+		memset(ibuf, 0, sizeof ibuf);
+		memset(obuf, 0, sizeof obuf);
 		break;
 
 	case MTA_AUTH_LOGIN_PASS:
-		bzero(ibuf, sizeof ibuf);
-		if (__b64_pton(s->relay->secret, (unsigned char *)ibuf, sizeof(ibuf)-1) == -1) {
+		memset(ibuf, 0, sizeof ibuf);
+		if (base64_decode(s->relay->secret, (unsigned char *)ibuf,
+				  sizeof(ibuf)-1) == -1) {
 			log_debug("debug: mta: %p: credentials too large on session", s);
 			mta_error(s, "Credentials too large");
 			break;
 		}
 
 		offset = strlen(ibuf+1)+2;
-		bzero(obuf, sizeof obuf);
-		__b64_ntop((unsigned char *)ibuf + offset, strlen(ibuf + offset), obuf, sizeof obuf);
+		memset(obuf, 0, sizeof obuf);
+		base64_encode((unsigned char *)ibuf + offset, strlen(ibuf + offset), obuf, sizeof obuf);
 		mta_send(s, "%s", obuf);
 
-		bzero(ibuf, sizeof ibuf);
-		bzero(obuf, sizeof obuf);
+		memset(ibuf, 0, sizeof ibuf);
+		memset(obuf, 0, sizeof obuf);
 		break;
 
 	case MTA_READY:
@@ -768,15 +766,38 @@ mta_enter_state(struct mta_session *s, int newstate)
 		break;
 
 	case MTA_MAIL:
+		s->currevp = TAILQ_FIRST(&s->task->envelopes);
+
+		e = s->currevp;
 		s->hangon = 0;
 		s->msgtried++;
-		mta_send(s, "MAIL FROM:<%s>", s->task->sender);
+		envid_sz = strlen(e->dsn_envid);
+		if (s->ext & MTA_EXT_DSN) {
+			mta_send(s, "MAIL FROM:<%s>%s%s%s%s",
+			    s->task->sender,
+			    e->dsn_ret ? " RET=" : "",
+			    e->dsn_ret ? dsn_strret(e->dsn_ret) : "",
+			    envid_sz ? " ENVID=" : "",
+			    envid_sz ? e->dsn_envid : "");
+		} else
+			mta_send(s, "MAIL FROM:<%s>", s->task->sender);
 		break;
 
 	case MTA_RCPT:
 		if (s->currevp == NULL)
 			s->currevp = TAILQ_FIRST(&s->task->envelopes);
-		mta_send(s, "RCPT TO:<%s>", s->currevp->dest);
+
+		e = s->currevp;
+		if (s->ext & MTA_EXT_DSN) {
+			mta_send(s, "RCPT TO:<%s>%s%s%s%s",
+			    e->dest,
+			    e->dsn_notify ? " NOTIFY=" : "",
+			    e->dsn_notify ? dsn_strnotify(e->dsn_notify) : "",
+			    e->dsn_orcpt ? " ORCPT=" : "",
+			    e->dsn_orcpt ? e->dsn_orcpt : "");
+		} else
+			mta_send(s, "RCPT TO:<%s>", e->dest);
+
 		s->rcptcount++;
 		break;
 
@@ -965,7 +986,7 @@ mta_response(struct mta_session *s, char *line)
 
 		s->currevp = TAILQ_NEXT(s->currevp, entry);
 		if (line[0] == '2') {
-			mta_flush_failedqueue(s);
+			s->failures = 0;
 			/*
 			 * this host is up, reschedule envelopes that
 			 * were cached for reschedule.
@@ -978,6 +999,7 @@ mta_response(struct mta_session *s, char *line)
 				delivery = IMSG_DELIVERY_PERMFAIL;
 			else
 				delivery = IMSG_DELIVERY_TEMPFAIL;
+			s->failures++;
 
 			/* remove failed envelope from task list */
 			TAILQ_REMOVE(&s->task->envelopes, e, entry);
@@ -1001,24 +1023,17 @@ mta_response(struct mta_session *s, char *line)
 				mta_delivery_log(e, sa_to_text(sa),
 				    buf, delivery, line);
 
-			/* push failed envelope to the session fail queue */
-			s->failed[s->failedcount] = e;
-			s->failedcount++;
+			if (domain)
+				mta_hoststat_update(domain, e->status);
+			mta_delivery_notify(e);
 
-			/*
-			 * if session fail queue is full:
-			 * - flush failed queue (failure w/ penalty)
-			 * - flush remaining tasks with TempFail
-			 * - mark route down
-			 */
-			if (s->failedcount == MAX_FAILED_ENVELOPES) {
-				mta_flush_failedqueue(s);
-				mta_flush_task(s, IMSG_DELIVERY_TEMPFAIL,
-				    "Host temporarily disabled", 0, 1);
-				mta_route_down(s->relay, s->route);
-				mta_enter_state(s, MTA_QUIT);
-				break;
-			}
+			if (s->relay->limits->max_failures_per_session &&
+			    s->failures == s->relay->limits->max_failures_per_session) {
+					mta_flush_task(s, IMSG_DELIVERY_TEMPFAIL,
+					    "Too many consecutive errors, closing connection", 0, 1);
+					mta_enter_state(s, MTA_QUIT);
+					break;
+				}
 
 			/*
 			 * if no more envelopes, flush failed queue
@@ -1038,7 +1053,6 @@ mta_response(struct mta_session *s, char *line)
 		break;
 
 	case MTA_DATA:
-		mta_flush_failedqueue(s);
 		if (line[0] == '2' || line[0] == '3') {
 			mta_enter_state(s, MTA_BODY);
 			break;
@@ -1083,7 +1097,6 @@ mta_response(struct mta_session *s, char *line)
 		break;
 
 	case MTA_RSET:
-		mta_flush_failedqueue(s);
 		s->rcptcount = 0;
 		if (s->relay->limits->sessdelay_transaction) {
 			log_debug("debug: mta: waiting for %llds after reset",
@@ -1195,6 +1208,8 @@ mta_io(struct io *io, int evt)
 			}
 			else if (strcmp(msg, "PIPELINING") == 0)
 				s->ext |= MTA_EXT_PIPELINING;
+			else if (strcmp(msg, "DSN") == 0)
+				s->ext |= MTA_EXT_DSN;
 		}
 
 		if (cont)
@@ -1350,6 +1365,7 @@ mta_flush_task(struct mta_session *s, int delivery, const char *error, size_t co
 
 		/* we're about to log, associate session to envelope */
 		e->session = s->id;
+		e->ext = s->ext;
 
 		/* XXX */
 		/*
@@ -1364,13 +1380,14 @@ mta_flush_task(struct mta_session *s, int delivery, const char *error, size_t co
 			mta_delivery_log(e, sa_to_text(sa),
 			    relay, delivery, error);
 
-		mta_delivery_notify(e, 0);
+		mta_delivery_notify(e);
 
 		domain = strchr(e->dest, '@');
 		if (domain) {
-			mta_hoststat_update(domain + 1, error);
+			domain++;
+			mta_hoststat_update(domain, error);
 			if (cache)
-				mta_hoststat_cache(domain + 1, e->id);
+				mta_hoststat_cache(domain, e->id);
 		}
 
 		n++;
@@ -1390,28 +1407,6 @@ mta_flush_task(struct mta_session *s, int delivery, const char *error, size_t co
 	stat_decrement("mta.task", 1);
 }
 
-static void
-mta_flush_failedqueue(struct mta_session *s)
-{
-	int			 i;
-	struct mta_envelope	*e;
-	const char		*domain;
-	uint32_t		 penalty;
-
-	penalty = s->failedcount == MAX_FAILED_ENVELOPES ? 1 : 0;
-	for (i = 0; i < s->failedcount; ++i) {
-		e = s->failed[i];
-
-		domain = strchr(e->dest, '@');
-		if (domain)
-			mta_hoststat_update(domain + 1, e->status);
-
-		mta_delivery_notify(e, penalty);
-	}
-
-	s->failedcount = 0;
-}
-
 static void
 mta_error(struct mta_session *s, const char *fmt, ...)
 {
@@ -1498,8 +1493,8 @@ mta_start_tls(struct mta_session *s)
 	struct ca_cert_req_msg	req_ca_cert;
 	const char	       *certname;
 
-	if (s->relay->cert)
-		certname = s->relay->cert;
+	if (s->relay->pki_name)
+		certname = s->relay->pki_name;
 	else
 		certname = s->helo;
 
@@ -1539,9 +1534,9 @@ mta_verify_certificate(struct mta_session *s)
 	s->flags |= MTA_WAIT;
 
 	/* Send the client certificate */
-	bzero(&req_ca_vrfy, sizeof req_ca_vrfy);
-	if (s->relay->cert)
-		pkiname = s->relay->cert;
+	memset(&req_ca_vrfy, 0, sizeof req_ca_vrfy);
+	if (s->relay->pki_name)
+		pkiname = s->relay->pki_name;
 	else
 		pkiname = s->helo;
 	if (strlcpy(req_ca_vrfy.pkiname, pkiname, sizeof req_ca_vrfy.pkiname)
@@ -1564,7 +1559,7 @@ mta_verify_certificate(struct mta_session *s)
 	if (xchain) {		
 		/* Send the chain, one cert at a time */
 		for (i = 0; i < sk_X509_num(xchain); ++i) {
-			bzero(&req_ca_vrfy, sizeof req_ca_vrfy);
+			memset(&req_ca_vrfy, 0, sizeof req_ca_vrfy);
 			req_ca_vrfy.reqid = s->id;
 			x = sk_X509_value(xchain, i);
 			req_ca_vrfy.cert_len = i2d_X509(x, &req_ca_vrfy.cert);
@@ -1579,7 +1574,7 @@ mta_verify_certificate(struct mta_session *s)
 	}
 
 	/* Tell lookup process that it can start verifying, we're done */
-	bzero(&req_ca_vrfy, sizeof req_ca_vrfy);
+	memset(&req_ca_vrfy, 0, sizeof req_ca_vrfy);
 	req_ca_vrfy.reqid = s->id;
 	m_compose(p_lka, IMSG_LKA_SSL_VERIFY, 0, 0, -1,
 	    &req_ca_vrfy, sizeof req_ca_vrfy);
@@ -1587,6 +1582,44 @@ mta_verify_certificate(struct mta_session *s)
 	return 1;
 }
 
+static const char *
+dsn_strret(enum dsn_ret ret)
+{
+	if (ret == DSN_RETHDRS)
+		return "HDRS";
+	else if (ret == DSN_RETFULL)
+		return "FULL";
+	else {
+		log_debug("mta: invalid ret %d", ret);
+		return "???";
+	}
+}
+
+static const char *
+dsn_strnotify(uint8_t arg)
+{
+	static char	buf[32];
+	size_t		sz;
+
+	if (arg & DSN_SUCCESS)
+		strlcat(buf, "SUCCESS,", sizeof(buf));
+
+	if (arg & DSN_FAILURE)
+		strlcat(buf, "FAILURE,", sizeof(buf));
+
+	if (arg & DSN_DELAY)
+		strlcat(buf, "DELAY,", sizeof(buf));
+
+	if (arg & DSN_NEVER)
+		strlcat(buf, "NEVER,", sizeof(buf));
+
+	/* trim trailing comma */
+	sz = strlen(buf);
+	if (sz)
+		buf[sz - 1] = '\0';
+
+	return (buf);
+}
 
 #define CASE(x) case x : return #x
 
