@@ -361,7 +361,7 @@ queue_fs_envelope_walk(uint64_t *evpid, char *buf, size_t len)
 		hdl = fsqueue_qwalk_new();
 
 	if (fsqueue_qwalk(hdl, evpid)) {
-		bzero(buf, len);
+		memset(buf, 0, len);
 		r = queue_fs_envelope_load(*evpid, buf, len);
 		if (r) {
 			msgid = evpid_to_msgid(*evpid);
