@@ -474,12 +474,12 @@ scheduler_ram_batch(int typemask, struct scheduler_batch *ret)
 			ret->delay = evp->sched - currtime;
 		else
 			ret->delay = evp->expire - currtime;
-		return (1);
+		goto done;
 	}
 	else {
 		ret->type = SCHED_NONE;
 		ret->evpcount = 0;
-		return (0);
+		goto done;
 	}
 
 	for (n = 0; (evp = TAILQ_FIRST(q)) && n < ret->evpcount; n++) {
@@ -516,7 +516,25 @@ scheduler_ram_batch(int typemask, struct scheduler_batch *ret)
 
 	ret->evpcount = n;
 
-	return (1);
+   done:
+
+	ret->mask = 0;
+	if (TAILQ_FIRST(&ramqueue.q_removed))
+		ret->mask |= SCHED_REMOVE;
+	if (TAILQ_FIRST(&ramqueue.q_expired))
+		ret->mask |= SCHED_EXPIRE;
+	if (TAILQ_FIRST(&ramqueue.q_update))
+		ret->mask |= SCHED_UPDATE;
+	if (TAILQ_FIRST(&ramqueue.q_bounce))
+		ret->mask |= SCHED_BOUNCE;
+	if (TAILQ_FIRST(&ramqueue.q_mda))
+		ret->mask |= SCHED_MDA;
+	if (TAILQ_FIRST(&ramqueue.q_mta))
+		ret->mask |= SCHED_MTA;
+	if (TAILQ_FIRST(&ramqueue.q_pending))
+		ret->mask |= SCHED_DELAY;
+
+	return ((ret->type == SCHED_NONE) ? 0 : 1);
 }
 
 static size_t
@@ -743,7 +761,7 @@ sorted_insert(struct rq_queue *rq, struct rq_envelope *evp)
 static void
 rq_queue_init(struct rq_queue *rq)
 {
-	bzero(rq, sizeof *rq);
+	memset(rq, 0, sizeof *rq);
 	tree_init(&rq->messages);
 	TAILQ_INIT(&rq->q_pending);
 	TAILQ_INIT(&rq->q_inflight);
