@@ -102,7 +102,7 @@ static int	offline_enqueue(char *);
 static void	purge_task(int, short, void *);
 static void	log_imsg(int, int, struct imsg *);
 static int	parent_auth_user(const char *, const char *);
-static void	load_ssl_tree(void);
+static void	load_pki_tree(void);
 
 enum child_type {
 	CHILD_DAEMON,
@@ -325,8 +325,8 @@ usage(void)
 {
 	extern char	*__progname;
 
-	fprintf(stderr, "usage: %s [-dnv] [-D macro=value] "
-	    "[-f file] [-P system]\n", __progname);
+	fprintf(stderr, "usage: %s [-dhnv] [-D macro=value] "
+	    "[-f file] [-P system] [-T trace]\n", __progname);
 	exit(1);
 }
 
@@ -358,53 +358,14 @@ parent_send_config(int fd, short event, void *p)
 	parent_send_config_lka();
 	parent_send_config_mfa();
 	parent_send_config_smtp();
-	purge_config(PURGE_SSL);
+	purge_config(PURGE_PKI);
 }
 
 static void
 parent_send_config_smtp(void)
 {
-	struct listener		*l;
-	struct ssl		*s;
-	void			*iter = NULL;
-	struct iovec		 iov[5];
-	int			 opt;
-
 	log_debug("debug: parent_send_config: configuring smtp");
 	m_compose(p_smtp, IMSG_CONF_START, 0, 0, -1, NULL, 0);
-
-	while (dict_iter(env->sc_ssl_dict, &iter, NULL, (void **)&s)) {
-		iov[0].iov_base = s;
-		iov[0].iov_len = sizeof(*s);
-		iov[1].iov_base = s->ssl_cert;
-		iov[1].iov_len = s->ssl_cert_len;
-		iov[2].iov_base = s->ssl_key;
-		iov[2].iov_len = s->ssl_key_len;
-		iov[3].iov_base = s->ssl_dhparams;
-		iov[3].iov_len = s->ssl_dhparams_len;
-		iov[4].iov_base = s->ssl_ca;
-		iov[4].iov_len = s->ssl_ca_len;
-		m_composev(p_smtp, IMSG_CONF_SSL, 0, 0, -1, iov, nitems(iov));
-	}
-
-	TAILQ_FOREACH(l, env->sc_listeners, entry) {
-		if ((l->fd = socket(l->ss.ss_family, SOCK_STREAM, 0)) == -1)
-			fatal("smtpd: socket");
-		opt = 1;
-#ifdef SO_REUSEADDR
-		if (setsockopt(l->fd, SOL_SOCKET, SO_REUSEADDR, &opt,
-			sizeof(opt)) < 0)
-#else
-		if (setsockopt(l->fd, SOL_SOCKET, SO_REUSEPORT, &opt,
-			sizeof(opt)) < 0)
-#endif
-			fatal("smtpd: setsockopt");
-		if (bind(l->fd, (struct sockaddr *)&l->ss, SS_LEN(&l->ss)) == -1)
-			fatal("smtpd: bind");
-		m_compose(p_smtp, IMSG_CONF_LISTENER, 0, 0, l->fd,
-		    l, sizeof(*l));
-	}
-
 	m_compose(p_smtp, IMSG_CONF_END, 0, 0, -1, NULL, 0);
 }
 
@@ -426,93 +387,8 @@ parent_send_config_mfa()
 void
 parent_send_config_lka()
 {
-	struct rule	       *r;
-	struct table	       *t;
-	void		       *iter_tree;
-	void		       *iter_dict;
-	const char	       *k;
-	char		       *v;
-	char		       *buffer;
-	size_t			buflen;
-	struct ssl	       *s;
-	struct iovec		iov[5];
-
 	log_debug("debug: parent_send_config_ruleset: reloading");
 	m_compose(p_lka, IMSG_CONF_START, 0, 0, -1, NULL, 0);
-
-	iter_dict = NULL;
-	while (dict_iter(env->sc_ssl_dict, &iter_dict, NULL, (void **)&s)) {
-		iov[0].iov_base = s;
-		iov[0].iov_len = sizeof(*s);
-		iov[1].iov_base = s->ssl_cert;
-		iov[1].iov_len = s->ssl_cert_len;
-		iov[2].iov_base = s->ssl_key;
-		iov[2].iov_len = s->ssl_key_len;
-		iov[3].iov_base = s->ssl_dhparams;
-		iov[3].iov_len = s->ssl_dhparams_len;
-		iov[4].iov_base = s->ssl_ca;
-		iov[4].iov_len = s->ssl_ca_len;
-		m_composev(p_lka, IMSG_CONF_SSL, 0, 0, -1, iov, nitems(iov));
-	}
-
-	iter_tree = NULL;
-	while (dict_iter(env->sc_tables_dict, &iter_tree, NULL,
-		(void **)&t)) {
-		m_compose(p_lka, IMSG_CONF_TABLE, 0, 0, -1, t, sizeof(*t));
-
-		iter_dict = NULL;
-		while (dict_iter(&t->t_dict, &iter_dict, &k,
-			(void **)&v)) {
-			buflen = strlen(k) + 1;
-			if (v)
-				buflen += strlen(v) + 1;
-			buffer = xcalloc(1, buflen,
-			    "parent_send_config_ruleset");
-			memcpy(buffer, k, strlen(k) + 1);
-			if (v)
-				memcpy(buffer + strlen(k) + 1, v,
-				    strlen(v) + 1);
-			m_compose(p_lka, IMSG_CONF_TABLE_CONTENT, 0, 0, -1,
-			    buffer, buflen);
-			free(buffer);
-		}
-	}
-
-	TAILQ_FOREACH(r, env->sc_rules, r_entry) {
-		m_compose(p_lka, IMSG_CONF_RULE, 0, 0, -1, r, sizeof(*r));
-		m_compose(p_lka, IMSG_CONF_RULE_SOURCE, 0, 0, -1,
-		    &r->r_sources->t_name,
-		    sizeof(r->r_sources->t_name));
-		if (r->r_senders) {
-			m_compose(p_lka, IMSG_CONF_RULE_SENDER,
-			    0, 0, -1,
-			    &r->r_senders->t_name,
-			    sizeof(r->r_senders->t_name));
-		}
-		if (r->r_recipients) {
-			m_compose(p_lka, IMSG_CONF_RULE_RECIPIENT,
-			    0, 0, -1,
-			    &r->r_recipients->t_name,
-			    sizeof(r->r_recipients->t_name));
-		}
-		if (r->r_destination) {
-			m_compose(p_lka, IMSG_CONF_RULE_DESTINATION,
-			    0, 0, -1,
-			    &r->r_destination->t_name,
-			    sizeof(r->r_destination->t_name));
-		}
-		if (r->r_mapping) {
-			m_compose(p_lka, IMSG_CONF_RULE_MAPPING, 0, 0, -1,
-			    &r->r_mapping->t_name,
-			    sizeof(r->r_mapping->t_name));
-		}
-		if (r->r_userbase) {
-			m_compose(p_lka, IMSG_CONF_RULE_USERS, 0, 0, -1,
-			    &r->r_userbase->t_name,
-			    sizeof(r->r_userbase->t_name));
-		}
-	}
-
 	m_compose(p_lka, IMSG_CONF_END, 0, 0, -1, NULL, 0);
 }
 
@@ -657,7 +533,7 @@ main(int argc, char *argv[])
 
 	TAILQ_INIT(&offline_q);
 
-	while ((c = getopt(argc, argv, "B:dD:nP:f:T:v")) != -1) {
+	while ((c = getopt(argc, argv, "B:dD:hnP:f:T:v")) != -1) {
 		switch (c) {
 		case 'B':
 			if (strstr(optarg, "queue=") == optarg)
@@ -680,6 +556,10 @@ main(int argc, char *argv[])
 				    "could not parse macro definition %s",
 				    optarg);
 			break;
+		case 'h':
+			log_info("version: OpenSMTPD " SMTPD_VERSION);
+			usage();
+			break;
 		case 'n':
 			debug = 2;
 			opts |= SMTPD_OPT_NOACTION;
@@ -768,7 +648,7 @@ main(int argc, char *argv[])
 		errx(1, "config file exceeds SMTPD_MAXPATHLEN");
 
 	if (env->sc_opts & SMTPD_OPT_NOACTION) {
-		load_ssl_tree();
+		load_pki_tree();
 		fprintf(stderr, "configuration OK\n");
 		exit(0);
 	}
@@ -795,7 +675,7 @@ main(int argc, char *argv[])
 	log_init(foreground);
 	log_verbose(verbose);
 
-	load_ssl_tree();
+	load_pki_tree();
 
 	log_info("info: %s %s starting", SMTPD_NAME, SMTPD_VERSION);
 
@@ -846,7 +726,7 @@ main(int argc, char *argv[])
 	config_done();
 
 	evtimer_set(&config_ev, parent_send_config, NULL);
-	bzero(&tv, sizeof(tv));
+	memset(&tv, 0, sizeof(tv));
 	evtimer_add(&config_ev, &tv);
 
 	/* defer offline scanning for a second */
@@ -873,32 +753,32 @@ main(int argc, char *argv[])
 }
 
 static void
-load_ssl_tree(void)
+load_pki_tree(void)
 {
-	struct ssl	*ssl;
-	void		*iter_dict;
+	struct pki	*pki;
 	const char	*k;
+	void		*iter_dict;
 
 	log_debug("debug: init ssl-tree");
 	iter_dict = NULL;
-	while (dict_iter(env->sc_ssl_dict, &iter_dict, &k, (void **)&ssl)) {
+	while (dict_iter(env->sc_pki_dict, &iter_dict, &k, (void **)&pki)) {
 		log_debug("info: loading pki information for %s", k);
-		if (ssl->ssl_cert_file == NULL)
-			fatalx("load_ssl_tree: missing certificate file");
-		if (ssl->ssl_key_file == NULL)
-			fatalx("load_ssl_tree: missing key file");
+		if (pki->pki_cert_file == NULL)
+			fatalx("load_pki_tree: missing certificate file");
+		if (pki->pki_key_file == NULL)
+			fatalx("load_pki_tree: missing key file");
 
-		if (! ssl_load_certificate(ssl, ssl->ssl_cert_file))
-			fatalx("load_ssl_tree: failed to load certificate file");
-		if (! ssl_load_keyfile(ssl, ssl->ssl_key_file, k))
-			fatalx("load_ssl_tree: failed to load key file");
+		if (! ssl_load_certificate(pki, pki->pki_cert_file))
+			fatalx("load_pki_tree: failed to load certificate file");
+		if (! ssl_load_keyfile(pki, pki->pki_key_file, k))
+			fatalx("load_pki_tree: failed to load key file");
 
-		if (ssl->ssl_ca_file)
-			if (! ssl_load_cafile(ssl, ssl->ssl_ca_file))
-				fatalx("load_ssl_tree: failed to load CA file");
-		if (ssl->ssl_dhparams_file)
-			if (! ssl_load_dhparams(ssl, ssl->ssl_dhparams_file))
-				fatalx("load_ssl_tree: failed to load dhparams file");
+		if (pki->pki_ca_file)
+			if (! ssl_load_cafile(pki, pki->pki_ca_file))
+				fatalx("load_pki_tree: failed to load CA file");
+		if (pki->pki_dhparams_file)
+			if (! ssl_load_dhparams(pki, pki->pki_dhparams_file))
+				fatalx("load_pki_tree: failed to load dhparams file");
 	}
 }
 
@@ -907,21 +787,6 @@ fork_peers(void)
 {
 	tree_init(&children);
 
-	/*
-	 * Pick descriptor limit that will guarantee impossibility of fd
-	 * starvation condition.  The logic:
-	 *
-	 * Treat hardlimit as 100%.
-	 * Limit smtp to 50% (inbound connections)
-	 * Limit mta to 50% (outbound connections)
-	 * Limit mda to 50% (local deliveries)
-	 * In all three above, compute max session limit by halving the fd
-	 * limit (50% -> 25%), because each session costs two fds.
-	 * Limit queue to 100% to cover the extreme case when tons of fds are
-	 * opened for all four possible purposes (smtp, mta, mda, bounce)
-	 */
-	fdlimit(0.5);
-
 	init_pipes();
 
 	child_add(queue(), CHILD_DAEMON, proc_title(PROC_QUEUE));
@@ -1205,7 +1070,7 @@ offline_enqueue(char *name)
 		size_t	 len;
 		arglist	 args;
 
-		bzero(&args, sizeof(args));
+		memset(&args, 0, sizeof(args));
 
 		if (lstat(path, &sb) == -1) {
 			log_warn("warn: smtpd: lstat: %s", path);
@@ -1513,6 +1378,7 @@ imsg_to_str(int type)
 	CASE(IMSG_CTL_LIST_ENVELOPES);
 	CASE(IMSG_CTL_REMOVE);
 	CASE(IMSG_CTL_SCHEDULE);
+	CASE(IMSG_CTL_SHOW_STATUS);
 
 	CASE(IMSG_CTL_TRACE);
 	CASE(IMSG_CTL_UNTRACE);
@@ -1523,6 +1389,9 @@ imsg_to_str(int type)
 	CASE(IMSG_CTL_MTA_SHOW_RELAYS);
 	CASE(IMSG_CTL_MTA_SHOW_ROUTES);
 	CASE(IMSG_CTL_MTA_SHOW_HOSTSTATS);
+	CASE(IMSG_CTL_MTA_BLOCK);
+	CASE(IMSG_CTL_MTA_UNBLOCK);
+	CASE(IMSG_CTL_MTA_SHOW_BLOCK);
 
 	CASE(IMSG_CONF_START);
 	CASE(IMSG_CONF_SSL);
@@ -1659,8 +1528,8 @@ parent_auth_pam(const char *username, const char *password)
 	int rc;
 	pam_handle_t *pamh = NULL;
 	struct pam_conv conv = { pam_conv_password, password };
-	
-	if ((rc = pam_start("smtpd", username, &conv, &pamh)) != PAM_SUCCESS)
+
+	if ((rc = pam_start(USE_PAM_SERVICE, username, &conv, &pamh)) != PAM_SUCCESS)
 		goto end;
 	if ((rc = pam_authenticate(pamh, 0)) != PAM_SUCCESS)
 		goto end;
